==> pb_common-0.63.tar.gz: pb_common-0.63/pb_common/services/bender_provider.py <==
import json
import time
import uuid
from typing import List

from urllib.parse import quote_plus

from attributedict.collections import AttributeDict
from dateutil.parser import parse

from pb_common.utils.logger import get_logger

__author__ = "Maksim Senchuk"

logger = get_logger(__name__)


class ClientRegistryError(Exception):
    pass


class ClientProviderError(Exception):
    pass


class ClientError(Exception):
    pass


class AgreementError(Exception):
    pass


class Card(AttributeDict):
    """
    Карта договора
    """

    def __init__(self, context, account_id, card):
        super().__init__(card)
        self._context = context
        self._account_id = account_id
        self._card_no = ''

    def __call__(self, status='open', **kwargs) -> List:
        """
        Найти карты по параметрам
        :param status: статус карты (из Контракт Сервиса)
        :param kwargs: параметры
        :return: [Card, ..., ]
        """
        return self._find_cards(status, **kwargs)

    def _find_cards(self, status: str, **kwargs) -> List:
        """
        Найти карты по параметрам
        :param status: статус карты (из Контракт Сервиса)
        :param kwargs: параметры
        :return: [Card, ..., ]
        """
        founds = []
        for card in self:
            if card.status == status:
                for key, value in kwargs.items():
                    if getattr(card, key, None) != value:
                        break
                else:
                    founds.append(card)
        if founds:
            return founds
        raise AgreementError('Не найдено карт с параметрами: {}'.format(kwargs))

    @property
    def card_no(self):
        """
        Номер карты - не маскированный
        """
        if self._card_no != '':
            return self._card_no

        v_cards = self._context.vabank_provider.get_cards(self._account_id)
        for v_card in v_cards:
            if getattr(self, 'cardNo', None):
                self._card_no = None
            if v_card['CARDNO'][-4:] == self.cardNo[-4:]:
                self._card_no = v_card['CARDNO']
        return self._card_no

    @property
    def quartile(self):
        """
        Третья квартиль карты
        """
        if self.card_no:
            return self.card_no[8:12]


class Cards(list):
    """
    Список карт договора
    """

    def __init__(self, context, agreement_id, cards: List):
        self._context = context
        self._agreement_id = agreement_id
        super().__init__(cards)

    def __getattr__(self, item):
        """
        Получить данные основной карты к договору
        :param item: может быть любой тэг из блока cards в ответе ContractService
        """
        for card in self:
            if card.primary is True and card.status == 'open':
                return getattr(card, item, None)
        raise AgreementError('Нет открытой основной карты')


class Agreement(AttributeDict):
    """
    Структура - договор клиента
    """

    def __init__(self, context, entries):
        super().__init__(entries)
        self._context = context
        self.cards = Cards(context, self.account_id, [Card(context, self.account_id, card) for card in self.cards])

    @property
    def productname(self):
        """
        Название продукта
        """
        return self.productName

    @property
    def account_id(self):
        """
        Номер договора
        """
        return self.agreementId

    @property
    def mk_auth_acct_no(self):
        """
        Номер счета договора в Мультикарте
        """
        return self.mkAccount

    @property
    def agr_prod_type(self):
        """
        Макропродукт договора
        """
        return self.macroProdType


class Agreements(list):
    """
    Список договоров клиента
    """

    def __init__(self, context, integration_id, agreements_bender: List = None):
        self._context = context
        self.integration_id = integration_id
        self._quartile_card = None
        self._agreements = self.get_contracts()

        # Проверить бьются ли договора подготовленные бендером с договорами клиента, которые вернул ContractService
        # если данные совпадают - клиент рабочий, если отличаются - клиент сломанный (возможно перезалили Вабанк),
        # нужно сгенерировать нового клиента
        if agreements_bender:
            for agr in agreements_bender:
                for agreement in self._agreements:
                    if str(agr['account_id']) == agreement['agreementId']:
                        if agr['mk_auth_acct_no'] == agreement['mkAccount']:
                            break
                        else:
                            raise ClientError('В ответе ContractService отличаются параметры '
                                                         'подготовленного договора: {}'.format(agr['account_id']))
                else:
                    raise ClientError(
                        'В ответе ContractService не найден подготовленный договор: {}'.format(agr['account_id']))

        super().__init__(self._agreements)

    def _find_agreements(self, status: str, **kwargs) -> List:
        """
        Найти договоры по параметрам
        :param status: статус договора (из Контракт Сервиса)
        :param kwargs: параметры
        :return: [Agreement, ..., ]
        """
        founds = []
        for agree in self:
            if agree.status == status:
                for key, value in kwargs.items():
                    if getattr(agree, key, None) != value:
                        break
                else:
                    founds.append(agree)
        if founds:
            return founds
        raise AgreementError('Не найдено договоров с параметрами: {}'.format(kwargs))

    def __call__(self, status='opened', **kwargs) -> List:
        """
        Найти договоры по параметрам
        :param status: статус договора (из Контракт Сервиса)
        :param kwargs: параметры
        :return: [Agreement]
        """
        return self._find_agreements(status, **kwargs)

    @property
    def curacc(self):
        """
        Получить договор Сбер счета
        """
        return self._find_agreements(status='opened', agr_prod_type='VTB_CURACC')[0]

    def get_contracts(self):
        """
        Получить контракты клиента из ContractService методом GET /internal/api/v2/contracts
        :return:
        """
        r = self._context.contract_service.get_contracts(
            path=dict(isOnline='true'),
            headers=dict(Authorization='IntegrationId {}'.format(self.integration_id))
        )
        if r.status_code == 200:
            return [Agreement(self._context, agreement) for agreement in r.json().get('agreements', [])]
        else:
            raise ClientError(
                'Не удалось получить данные по договорам клиента в Contract Service: {}'.format(r.text))

    @property
    def quartile_card(self):
        """
        Получить 3-ю квартиль карты для регистрации в ИБ
        :return: 3-я квартиль карты
        """
        if self._quartile_card:
            return self._quartile_card

        for agree in self._agreements:
            if agree['status'] == 'opened' and agree['cards'] and agree['macroProdType'] in ['VTB_CURACC', 'VTB_CARD']:
                for card in agree['cards']:
                    if card['status'] == 'open':
                        v_cards = self._context.vabank_provider.get_cards(agree['agreementId'])
                        for v_card in v_cards:
                            if v_card['CARDNO'][-4:] == card['cardNo'][-4:]:
                                self._quartile_card = v_card['CARDNO'][8:12]
                                return self._quartile_card
        raise AgreementError('Не найдено открытых договоров с картами')


class Client(object):
    """
    Объект - клиент
    """

    def __init__(self, context, client, agreements):
        """
        Инициализация Клиента
        :param context: WebContext
        :param client: клиентские данные из Бендера
        :param agreements: список договоров из Бендера
        """
        self._context = context
        self._client = client
        self._client_info = self.get_client()
        self.agreements = Agreements(context, self.integration_id, agreements)
        self.registration = ClientRegistration(context.token_provider_service)

    def __getattr__(self, item):
        """
        Получить атрибут клиента, который вернул Бендер
        :param item:
        :return:
        """
        try:
            return self._client[item]
        except KeyError:
            raise ClientError('Не найден атрибут клиента {} в списке доступных: {}'.format(
                item, self._client.keys()))

    def get_client(self):
        """
        Получить информацию по клиенту из ClientService методом GET internal/api/v2/client/{valueId}?type={typeName}
        :return:
        """
        r = self._context.client_service.get_client(path=dict(valueId=self.integration_id, typeName='integrationId'))
        if r.status_code == 200:
            return r.json()
        else:
            raise ClientError('Не удалось получить данные по клиенту в Client Service: {}'.format(r.text))

    @staticmethod
    def _check_data(ref, exc, title):
        """Сравнить данные из Бенедера и сервиса"""
        if ref != exc:
            raise ClientError('Не совпадают данные подготовленного клиента и ответа сервиса. '
                              'Ожидаемый {} - {} , вернул сервис - {}'.format(title, ref, exc))

    @property
    def integration_id(self):
        """
        Вернуть integration_id клиента из данных полученных от Бендера
        :return: integration_id
        """
        try:
            return self._client['integration_id']
        except KeyError:
            ClientError('Не найден integration_id клиента')

    @property
    def mobile(self):
        """
        Вернуть номер мобильного телефона, при этом происходит проверка, что бендер и сервис вернули одинаковый номер.
        Если номер отличается, значит подготовленный клиент испорчен - возбуждается исключение ClientError
        :return: номер в формате +71234567890
        """
        at = self._client.get('mobile')
        srv = self._client_info.get('contact', {}).get('phoneNumber')
        self._check_data(at, srv, 'Номер телефона')
        return at

    @property
    def access_code(self):
        """
        Вернуть код доступа (для регистрации), при этом происходит проверка, что бендер и сервис вернули
        одинаковое значение.
        Если код доступа отличается, значит подготовленный клиент испорчен - возбуждается исключение ClientError
        :return: код доступа
        """
        at = self._client.get('access_code')
        srv = self._client_info.get('accessCode')
        self._check_data(at, srv, 'Access Code')
        return at

    @property
    def birth_date(self):
        """
        Вернуть дату рождения, при этом происходит проверка, что бендер и сервис вернули одинаковое значение.
        Если дата рождения отличается, значит подготовленный клиент испорчен - возбуждается исключение ClientError
        :return: дата рождения
        """
        srv = self._client_info.get('contact', {}).get('birthDate')
        at = self._client.get('birth_date')
        self._check_data(at, srv, 'дата рождения')
        if srv and at:
            self._check_data(parse(at), parse(srv), 'дата рождения')
            return at
        return None

    @property
    def name_f(self):
        """
        Получить имя фамилию клиента в формате: Имя Ф
        :return: Имя Ф
        """
        if self.first_name is not None and self.last_name is not None:
            return '{} {}.'.format(self.first_name.strip(), self.last_name.strip()[0]).title()
        return ''

    def get_birth_date(self, format_date='%d.%m.%Y'):
        """
        Получить дату рождения клиента в определнном формате
        :param format_date: формат даты
        :return: дата рождения
        """
        if self.birth_date is not None:
            birth_dt = parse(self.birth_date)
            return birth_dt.strftime(format_date)


class ClientProvider(object):
    """Получение тестовых данных из Бендера, генерация новых данных"""

    def __init__(self, context):
        self._context = context
        self.user = {}

    def __call__(self, action='GET', task_id=None, alias=None, is_client=False, timeout=0, **kwargs):
        """
        Общий метод работы с тестовыми данными
        :param action: действие - соответствует вызываемому методу Бендера
        :param task_id: task_id клиента
        :param alias: alias клиента
        :param is_client: признак, используется для создания новых тестовых данных в Бендере, если True - создан клиент
        :param timeout: время в секундах в течение которого, происходит опрос Бендера - подготовлены ли тестовые данные
        :param kwargs: параметры для создания новых тестовых данных
        """
        if action.upper() == 'GET':
            self.user = self._get(task_id, alias)
            return Client(self._context, **self.user['response'])
        elif action.upper() == 'RELEASE':
            return self._release()
        elif action.upper() == 'RERUN':
            return self._rerun()
        elif action.upper() == 'CREATE_TASK':
            self._create_task(**kwargs)
            start_time = time.time()
            while time.time() - start_time <= timeout:
                r = self._get_task()
                if r['state'] == 'COMPLETED':
                    if is_client:
                        self.user = r
                        return Client(self._context, **self.user['response'])
                    return r['response']
                delta = timeout - (time.time() - start_time)
                sleep = int(delta / (int(timeout / 10))) if int(delta / (int(timeout / 10))) >= 1 else 1
                logger.debug('Ожидать следующей проверки статуса задачи: {} сек.'.format(sleep))
                time.sleep(sleep)
            if timeout > 0:
                raise ClientProviderError('Не получен финальный статус подготовки тестовых данных '
                                                     'за {} секунд по задаче {}'.format(timeout, self.user['task_id']))
        elif action.upper() == 'GET_TASK':
            return self._get_task(**kwargs)
        elif action.upper() == 'DELETE':
            return self._delete()
        else:
            raise ClientProviderError('Неизвестное действие: {}'.format(action))

    def _get(self, task_id, alias):
        """
        Получить клиента из хранилища тестовых пользователей для автотестов.
        При этом если клиента нельзя использовать параллельно в нескольких тестах - клиенту будет
        установлен признак "занят" и обновится время последнего использования клиента.
        После выполнения теста - клиента нужно освободить методом _release.
        Если клиента можно использовать только один раз - нужно вызвать метод _delete после теста.
        :param task_id: идентификатор задачи в Бендере для получения конкретного клиента
        :param alias: алиас клиента, для получения любого свободного клиента по этому алиасу
        :return: задача
        """
        if task_id:
            r = self._context.bender_hub.acquire_client(
                path=dict(item='task_id'), params=dict(value=task_id, env=self._context.env))
        elif alias:
            r = self._context.bender_hub.acquire_client(
                path=dict(item='alias'), params=dict(value=alias, env=self._context.env))
        else:
            raise AttributeError('Не передан task_id или alias')
        if r.status_code == 200:
            return r.json()
        raise ClientProviderError('Ошибка при поиске клиента: {}'.format(r.text))

    def _release(self):
        """
        Освободить клиента в хранилище пользователей для автотестов.
        Отправляет запрос в Бендер, чтобы снять метку - "занят" с клиента.
        :return: True - успех, False - что-то пошло не так
        """
        r = self._context.bender_hub.release_client(path=dict(client_uuid=self.user['task_id']))
        logger.debug("Release client: {}".format(r.text))
        return r.status_code == 200

    def _delete(self):
        """
        Пометить клиента удаленным в хранилище пользователей для автотестов.
        :return: True - успех, False - что-то пошло не так
        """
        r = self._context.bender_hub.delete_client(path=dict(client_uuid=self.user['task_id']))
        logger.debug("Delete client: {}".format(r.text))
        return r.status_code == 200

    def _rerun(self):
        """
        Запустить на выполнение новую задачу, такую же как имеющаяся
        :return: True - успех, False - что-то пошло не так
        """
        r = self._context.bender_hub.re_create_client(path=dict(client_uuid=self.user['task_id']))
        return r.status_code == 200

    def _create_task(self, **kwargs):
        """
        Запустить на выполнение новую задачу
        """
        kwargs.update(dict(env=self._context.env))
        r = self._context.bender_hub.create_task(data=json.dumps(kwargs))
        if r.status_code == 200:
            self.user.update(r.json())
        else:
            raise ClientProviderError('Не удалось создать новую задачу: {}'.format(r.text))

    def _get_task(self, task_id=None):
        """
        Получить данные по задаче
        :param task_id: идентификатор задачи
        :return: задача
        """
        if not task_id:
            task_id = self.user['task_id']
        r = self._context.bender_hub.get_task(path=dict(client_uuid=task_id))
        if r.status_code == 200:
            return r.json()
        logger.warning(r.text)
        return {}


class ClientRegistration(object):
    """
    Регистрация в ДБО
    """

    def __init__(self, token_provider):
        self.token_provider = token_provider

    def create(self, integration_id, client_type, phone, password, has_password, login, has_login, name):
        """
        Зарегистрировать клиента в ДБО через TokenProvider
        :param integration_id: интегрейшн айди клиента
        :param client_type: тип клиента Full/Lite/LiteIdent
        :param phone: номер телефона клиента
        :param password: пароль
        :param has_password: признак наличия пароля True/False
        :param login: логин
        :param has_login: признак наличия логина True/False
        :param name: имя пользователя: Иванов Иван Иванович
        """
        session_info = {
            "integrationId": integration_id,
            "attributes": {
                "cn": login,
                "userPassword": password,
                "uid": str(uuid.uuid4()),
                "mobile": phone,
                "clientDisabled": "false",
                "clientType": client_type,
                "hasLogin": str(has_login).lower()
            }
        }
        if has_password is not None:
            session_info['attributes'].update(dict(hasPassword=str(has_password).lower()))
        if name:
            session_info['attributes'].update(dict(displayName=name))
        body = """client_id=factura&client_secret=ekassir&grant_type=sign_up&session_info={}""".format(
            quote_plus(json.dumps(session_info, ensure_ascii=False)))
        r = self.token_provider.auth_token(data=body)
        if r.status_code == 200:
            return r.json()
        raise ClientRegistryError('Клиент {} не зарегистрирован в ДБО'.format(integration_id))

    def delete(self, integration_id):
        """
        Удалить регистрацию клиента в ДБО через TokenProvider
        :param integration_id: интегрейшн айди клиента
        """
        session_info = {"integration_id": integration_id}
        body = """client_id=factura&client_secret=ekassir&grant_type=sign_out&session_info={}""".format(
            quote_plus(json.dumps(session_info, ensure_ascii=False)))
        r = self.token_provider.auth_token(data=body)
        logger.debug(r.text)

==> pb_common-0.96.tar.gz: pb_common-0.96/pb_common/services/bender_provider.py <==
import json
import time

from pb_common.client.client import Client
from pb_common.utils.logger import get_logger

__author__ = "Maksim Senchuk"

logger = get_logger(__name__)


class ClientProviderError(Exception):
    pass


class ClientProvider(object):
    """Получение тестовых данных из Бендера, генерация новых данных"""

    def __init__(self, context):
        self._context = context
        self.user = {}
        self._alias = None

    def __call__(self, action='GET', task_id=None, alias=None, is_client=False, timeout=0, **kwargs):
        """
        Общий метод работы с тестовыми данными
        :param action: действие - соответствует вызываемому методу Бендера
        :param task_id: task_id клиента
        :param alias: alias клиента
        :param is_client: признак, используется для создания новых тестовых данных в Бендере, если True - создан клиент
        :param timeout: время в секундах в течение которого, происходит опрос Бендера - подготовлены ли тестовые данные
        :param kwargs: параметры для создания новых тестовых данных
        """
        if action.upper() == 'GET':
            self.user = self._get(task_id, alias)
            return Client(self._context, **self.user['response'])
        elif action.upper() == 'GET_TEST_CLIENT':
            user = self._get(task_id, alias)
            self._release(user['task_id'])
            return Client(self._context, **user['response'])
        elif action.upper() == 'RELEASE':
            return self._release()
        elif action.upper() == 'RERUN':
            return self._rerun()
        elif action.upper() == 'CREATE_TASK':
            self._create_task(**kwargs)
            start_time = time.time()
            while time.time() - start_time <= timeout:
                r = self._get_task()
                if r['state'] == 'COMPLETED':
                    if is_client:
                        self.user = r
                        return Client(self._context, **self.user['response'])
                    return r['response']
                delta = timeout - (time.time() - start_time)
                sleep = int(delta / (int(timeout / 10))) if int(delta / (int(timeout / 10))) >= 1 else 1
                logger.debug('Ожидать следующей проверки статуса задачи: {} сек.'.format(sleep))
                time.sleep(sleep)
            if timeout > 0:
                raise ClientProviderError('Не получен финальный статус подготовки тестовых данных '
                                          'за {} секунд по задаче {}'.format(timeout, self.user['task_id']))
        elif action.upper() == 'GET_TASK':
            return self._get_task(**kwargs)
        elif action.upper() == 'DELETE':
            return self._delete()
        else:
            raise ClientProviderError('Неизвестное действие: {}'.format(action))

    @property
    def alias(self):
        """
        Получить alias тестового клиента
        :return: alias
        """
        return self._alias

    def _get(self, task_id, alias):
        """
        Получить клиента из хранилища тестовых пользователей для автотестов.
        При этом если клиента нельзя использовать параллельно в нескольких тестах - клиенту будет
        установлен признак "занят" и обновится время последнего использования клиента.
        После выполнения теста - клиента нужно освободить методом _release.
        Если клиента можно использовать только один раз - нужно вызвать метод _delete после теста.
        :param task_id: идентификатор задачи в Бендере для получения конкретного клиента
        :param alias: алиас клиента, для получения любого свободного клиента по этому алиасу
        :return: задача
        """
        if task_id:
            r = self._context.bender_hub.acquire_client(
                path=dict(item='task_id'), params=dict(value=task_id, env=self._context.env))
        elif alias:
            r = self._context.bender_hub.acquire_client(
                path=dict(item='alias'), params=dict(value=alias, env=self._context.env))
        else:
            raise AttributeError('Не передан task_id или alias')
        if r.status_code == 200:
            self._alias = alias
            return r.json()
        raise ClientProviderError('Ошибка при поиске клиента: {}'.format(r.text))

    def _release(self, task_id=None):
        """
        Освободить клиента в хранилище пользователей для автотестов.
        Отправляет запрос в Бендер, чтобы снять метку - "занят" с клиента.
        :return: True - успех, False - что-то пошло не так
        """
        if not task_id and self.user:
            task_id = self.user['task_id']
        elif not self.user and not task_id:
            logger.debug("Нет клиента, которого можно освободить")
            return False

        r = self._context.bender_hub.release_client(path=dict(client_uuid=task_id))
        self._alias = None
        logger.debug("Release client: {}".format(r.text))
        return r.status_code == 200

    def _delete(self):
        """
        Пометить клиента удаленным в хранилище пользователей для автотестов.
        :return: True - успех, False - что-то пошло не так
        """
        r = self._context.bender_hub.delete_client(path=dict(client_uuid=self.user['task_id']))
        self._alias = None
        logger.debug("Delete client: {}".format(r.text))
        return r.status_code == 200

    def _rerun(self):
        """
        Запустить на выполнение новую задачу, такую же как имеющаяся
        :return: True - успех, False - что-то пошло не так
        """
        r = self._context.bender_hub.re_create_client(path=dict(client_uuid=self.user['task_id']))
        return r.status_code == 200

    def _create_task(self, **kwargs):
        """
        Запустить на выполнение новую задачу
        """
        kwargs.update(dict(env=self._context.env))
        r = self._context.bender_hub.create_task(data=json.dumps(kwargs))
        if r.status_code == 200:
            self.user.update(r.json())
        else:
            raise ClientProviderError('Не удалось создать новую задачу: {}'.format(r.text))

    def _get_task(self, task_id=None):
        """
        Получить данные по задаче
        :param task_id: идентификатор задачи
        :return: задача
        """
        if not task_id:
            task_id = self.user['task_id']
        r = self._context.bender_hub.get_task(path=dict(client_uuid=task_id))
        if r.status_code == 200:
            return r.json()
        logger.warning(r.text)
        return {}
